#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import functools

from markdown import Markdown

from iambic import ast
from .markdown import render_markdown


EXTENSIONS = ("meta", "attr_list", "pymdownx.details", "tables", "pymdownx.tabbed")


@functools.lru_cache(maxsize=1)
def renderer(*extensions) -> Markdown:
    extensions = tuple(set(EXTENSIONS).union(set(extensions)))
    return Markdown(extensions=extensions)


def render_html(play: ast.Play, *extensions: str, table: bool = True) -> str:
    """Render a play instance as an html document

    This uses the `Python-Markdown` library for rendering the final html,
    so all extensions supported by that library are also supported with this function.

    Parameters
    ----------
    play
        The instance of `iambic.ast.Play`
    extensions
        Any additional Python-Markdown extensions.
    table : default True
        Optionally render a character mapping for your play.

    Returns
    -------
    str
        The HTML document.
    """
    text = render_markdown(play, table=table)
    return renderer(*extensions).convert(text)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import functools

import typic

from iambic import ast


class Parser:
    __resolver_map__ = ast.GenericNode.__resolver_map__

    @functools.lru_cache()
    def parse(self, data: str) -> ast.ResolvedNodeT:
        return typic.transmute(ast.ResolvedNodeT, data)  # type: ignore

    __call__ = parse


parser = Parser()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from textwrap import indent
from typing import Iterable, Mapping, Union, cast

from iambic import ast
from .table import RichMarker, tabulate, iter_tabs

TITLE = "# {0}"
ACT = "## {0}"
SCENE = "### {0}"
DIR = "*{0}*"
ACTION = r"*\[{0}]*"
CHAR = "**{0}**"
ID = "{{: id={0} }}"
LINK = '<a class="headerlink" href="#{0}" title="Permanent link">{1}</a>'


def iter_stage_direction(entry: Union[ast.Direction, ast.Entrance]) -> Iterable[str]:
    yield ""
    text = entry.action if isinstance(entry, ast.Direction) else entry.text
    line = DIR.format(text)
    link = LINK.format(entry.id, RichMarker.PRES.value)
    yield f"{line} {link}"
    yield ID.format(entry.id)
    yield ""


def iter_speech(speech: ast.Speech, persona: ast.Persona) -> Iterable[str]:
    character = CHAR.format(persona.name)
    charlink = LINK.format(speech.id, speech.linerange)
    yield f"{character} {charlink}  "
    last = None
    speech_linked = False
    for entry in speech.body:
        if isinstance(entry, ast.Action):
            line = ACTION.format(entry.action)
            # If this isn't the first line of the speech
            # and we haven't added an attribute to the first paragraph of the speech,
            # do so now.
            if last:
                if not speech_linked:
                    yield ID.format(speech.id)
                    speech_linked = True
                yield ""
            # yield the line with a link
            yield f"{line}  "
        elif isinstance(entry, (ast.Direction, ast.Entrance)):
            # Build a new linked paragraph for stage directions
            if last:
                if not speech_linked:
                    yield ID.format(speech.id)
                    speech_linked = True
                yield ""
            yield from iter_stage_direction(entry)
        else:
            # Otherwise continue building the speech.
            yield f"{entry.line}  "
        last = entry
    if not speech_linked:
        yield ID.format(speech.id)
    yield ""


def iter_scene(
    scene: ast.ActNodeT, personae: Mapping[ast.NodeID, ast.Persona], h1: bool = False
) -> Iterable[str]:
    if isinstance(scene, (ast.Prologue, ast.Epilogue)) and scene.as_act:
        yield from iter_act(scene, personae)
    heading = ACT.format(scene.text) if h1 else SCENE.format(scene.text)
    id = ID.format(scene.id)
    if isinstance(scene, ast.Intermission):
        yield f"{heading} {id}"
    else:
        setting = scene.setting or " "
        if setting.strip():
            setting = f". {setting} "
        yield f"{heading}{setting}{id}"
        yield ""
        body = cast(ast.SceneBodyT, scene.body)
        for entry in body:
            if isinstance(entry, (ast.Direction, ast.Entrance)):
                yield from iter_stage_direction(entry)
            else:
                persona = personae[entry.persona]
                yield from iter_speech(entry, persona)
    yield ""


def iter_act(
    act: ast.PlayNodeT, personae: Mapping[ast.NodeID, ast.Persona], h2: bool = False
) -> Iterable[str]:
    if isinstance(act, (ast.Prologue, ast.Epilogue)) and not act.as_act:
        yield from iter_scene(act, personae, h1=True)
    else:
        heading = SCENE.format(act.text) if h2 else ACT.format(act.text)
        id = ID.format(act.id)
        yield f"{heading} {id}"
        yield ""
        body = cast(ast.ActBodyT, act.body)
        for entry in body:
            yield from iter_scene(entry, personae)

    yield ""


def iter_table(play: ast.Play) -> Iterable[str]:
    tbl = tabulate(play, links=True, rich=True)
    yield ACT.format("Character Index")
    yield ""
    yield '???+ "Character Index"'
    yield ""
    for line in iter_tabs(tbl):
        yield indent(line, "    ")
    yield ""


def iter_play(play: ast.Play, table: bool = True) -> Iterable[str]:
    personae = {p.id: p for p in play.personae}
    if play.meta.title:
        yield f"# {play.meta.title}"
        yield ""
    if table:
        yield from iter_table(play)
    for act in play.body:
        yield from iter_act(act, personae)


def render_markdown(play: ast.Play, *, table: bool = True) -> str:
    """The renderer for a Markdown document.

    Parameters
    ----------
    play :
        The play tree, as loaded via :func:`~iambic.parse.text` or
        :func:`~iambic.parse.data`
    table : default True
        Optionally render a character mapping for your play.
    """
    return "\n".join(iter_play(play, table=table))



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import enum
import textwrap
from typing import List, Dict, Hashable, Tuple, Union, Type, Any, Iterable

import tablib

from iambic import ast


Row = List[Hashable]
Table = Dict[str, Row]
Matrix = List[Tuple[Hashable, ...]]


class Column(str, enum.Enum):
    """Predefined column names"""

    CHAR = "Dramatis Personae"
    APPR = "First Appearance"
    CLINE = "Lines"
    PLINE = "Player Lines"
    PLAYR = "Player"
    SORT = "Sort"
    GSWAP = " [SW]"


class Marker(str, enum.Enum):
    """The character to use when marking a persona as present in a scene."""

    SPEAK = "X"
    PRES = "O"
    NONE = ""


class RichMarker(str, enum.Enum):
    """The rich text character to use when marking a persona as present in a scene."""

    SPEAK = "💬"
    PRES = "👁️‍🗨️"
    NONE = ""


class Tabulator:
    """Generate a tabular representation of :py:class:`~iambic.ast.Play`

    Also known as a 'character map', this utility will generate a table
    which shows the number of lines for a given character and where they
    appear in the play.

    .. i.e.:
        | Dramatis Personae | Lines | I.i | I.ii | Player | Player Lines |
        ------------------------------------------------------------------
        | Johnny Appleseed  | 100   |  x  |   x  |  Jimmy |      75      |

    """

    def _tabulate_scene(
        self,
        scene: Union[ast.Scene, ast.Epilogue, ast.Prologue],
        node_column: List[str],
        cline_column: List[int],
        char_index: Dict[str, int],
        personae: Dict[str, ast.Persona],
        links: bool,
        marker_type: Type[Union[Marker, RichMarker]],
    ):
        for child in scene.body:
            marker = (
                marker_type.SPEAK
                if child.type == ast.NodeType.SPCH
                else marker_type.PRES
            )
            entry = self.link(marker, child) if links else marker.value
            if isinstance(child, ast.Speech):
                persona = personae[child.persona]
                index = char_index[persona.name]
                cline_column[index] += child.num_lines
                if marker_type.SPEAK not in node_column[index]:
                    node_column[index] = entry
            elif isinstance(child, ast.Entrance):
                for pers in child.personae:
                    persona = personae[pers]
                    index = char_index[persona.name]
                    if not node_column[index]:
                        node_column[index] = entry

    @staticmethod
    def link(marker: Union[Marker, RichMarker], node: ast.ResolvedNodeT) -> str:
        return f"[{marker.value}](#{node.id})"

    def tabulate(
        self, play: ast.Play, *, links: bool = False, rich: bool = False
    ) -> Table:
        """Generate a table for this play.

        The table is returned in the form of a Mapping.
        The keys are the header and the columns are the values.

        .. i.e.:
            {
                "foo": ['a', 'value'],
                "bar": ['another, 'value],
                ...
            }
        """
        marker_type: Type[Union[Marker, RichMarker]] = RichMarker if rich else Marker
        table: Dict[str, Any] = {
            Column.CHAR.value: [x.name for x in play.personae],
            Column.APPR.value: [x.index for x in play.personae],
            Column.CLINE.value: [0 for _ in play.personae],
        }
        char_column: List[str] = table[Column.CHAR.value]
        cline_column: List[int] = table[Column.CLINE.value]
        char_index: Dict[str, int] = {y: x for x, y in enumerate(char_column)}
        personae = {x.id: x for x in play.personae}
        for act in play.body:
            # Epilogues and Prologues can be shaped like Scenes or Acts.
            # And can be top-level, like Acts.
            children: ast.ActBodyT = act.body if (  # type: ignore
                isinstance(act, ast.Act) or act.as_act
            ) else (act,)
            for scene in children:

                table[scene.col] = [marker_type.NONE.value for _ in char_column]
                if isinstance(scene, ast.Intermission):
                    continue
                self._tabulate_scene(
                    scene=scene,
                    node_column=table[scene.col],
                    cline_column=cline_column,
                    char_index=char_index,
                    personae=personae,
                    links=links,
                    marker_type=marker_type,
                )

        return table

    @staticmethod
    def matrix(table: Table) -> Matrix:
        """Pivot a table into a 2-D array (matrix).

        The first entry is the header and the proceeding lines are values.

        .. i.e.:
            [
                ("foo", "bar"),
                ("a", "another"),
                ("value", "value"),
                ...
            ]
        """
        return [(*table.keys(),), *zip(*table.values())]

    def dataset(self, table: Table) -> tablib.Dataset:
        """Transform a table into an instance of :py:class:`tablib.Dataset`."""
        pivot = self.matrix(table)
        headers = pivot.pop(0)
        return tablib.Dataset(*pivot, headers=headers)

    def __call__(
        self, play: ast.Play, *, links: bool = False, rich: bool = False
    ) -> tablib.Dataset:
        table = self.tabulate(play, links=links, rich=rich)
        return self.dataset(table)


tabulate = Tabulator()


class TableFormat(str, enum.Enum):
    TABLE = "table"
    TABS = "tabs"


def iter_character_tab(table: tablib.Dataset) -> Iterable[str]:
    characters, lines = table[Column.CHAR], table[Column.CLINE]
    yield f'=== "{Column.CHAR.value}"'
    for character, line_count in zip(characters, lines):
        yield f"    - {character} ({line_count} lines)"
    yield ""


def iter_scene_tab(table: tablib.Dataset, scene_name: str) -> Iterable[str]:
    characters, scene = table[Column.CHAR], table[scene_name]
    yield f'=== "{scene_name}"'
    for character, marker in zip(characters, scene):
        if marker:
            yield f"    - {character} ⇒ {marker}"
    yield ""


def iter_grid_tab(table: tablib.Dataset) -> Iterable[str]:
    yield '=== "Full Breakdown"'
    yield textwrap.indent(export_grid(table), "    ")
    yield ""


def iter_tabs(table: tablib.Dataset, *, __headers=frozenset(Column)):
    yield from iter_character_tab(table)
    for header in (h for h in table.headers if h not in __headers):
        yield from iter_scene_tab(table, header)
    yield from iter_grid_tab(table)


def export_tabs(table: tablib.Dataset) -> str:
    return "\n".join(iter_tabs(table))


def export_grid(table: tablib.Dataset) -> str:
    return table.export("cli", tablefmt="github")


def export(table: tablib.Dataset, format: TableFormat = TableFormat.TABLE):
    if format == TableFormat.TABLE:
        return export_grid(table)
    return export_tabs(table)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import functools
from operator import attrgetter
from typing import (
    ClassVar,
    Optional,
    Union,
    Tuple,
    Any,
    Mapping,
    Type,
    List,
    TypeVar,
    NewType,
    Iterable,
)

import typic
from inflection import parameterize, titleize

from iambic import roman
from .base import NodeType, NodeMixin


__all__ = (
    "Act",
    "ActBodyT",
    "ActNodeT",
    "Scene",
    "SceneBodyT",
    "SceneNodeT",
    "Prologue",
    "Epilogue",
    "LogueBodyT",
    "Intermission",
    "Persona",
    "Entrance",
    "Exit",
    "Action",
    "Direction",
    "Dialogue",
    "Speech",
    "SpeechNodeT",
    "SpeechBodyT",
    "ResolvedNodeT",
    "Play",
    "PlayNodeT",
    "PlayBodyT",
    "Metadata",
    "GenericNode",
    "NodeID",
    "NodeType",
    "node_deserializer",
    "log_deserializer",
    "isnodetype",
)


T = TypeVar("T")
NodeID = NewType("NodeID", str)
indexgetter = attrgetter("index")


_T = TypeVar("_T")


def sort_body(body: Iterable[_T]) -> Tuple[_T, ...]:
    return tuple(sorted(body, key=indexgetter))


@typic.klass(unsafe_hash=True, order=True)
class Act(NodeMixin):
    """A representation of a single Act in a Play."""

    type: ClassVar[NodeType] = NodeType.ACT
    index: int
    text: str
    num: int
    body: "ActBodyT" = typic.field(compare=False, hash=False, default=())  # type: ignore

    @typic.cached_property
    def id(self) -> NodeID:
        return NodeID(parameterize(f"{self.type.lower()}-{self.col}"))

    @typic.cached_property
    def col(self):
        return roman.numeral(self.num)

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Act":
        numeral = node.pieces[1]
        num = int(numeral) if numeral.isdigit() else roman.integer(numeral)
        return cls(index=node.index, text=node.match_text, num=num)


@typic.klass(unsafe_hash=True)
class Scene(NodeMixin):
    """A representation of a single Scene in a play."""

    type: ClassVar[NodeType] = NodeType.SCENE
    index: int
    text: str
    num: int
    setting: Optional[str] = None
    act: Optional[NodeID] = None
    body: "SceneBodyT" = typic.field(compare=False, hash=False, default=())  # type: ignore
    personae: "PersonaeIDT" = typic.field(compare=False, hash=False, default=())  # type: ignore

    @typic.cached_property
    def id(self) -> NodeID:
        return NodeID(parameterize(f"{self.act}-{self.type.lower()}-{self.col}"))

    @typic.cached_property
    def col(self) -> str:
        pre = self.act or ""
        if pre:
            pieces = self.act.split("-")
            if len(pieces) == 2:
                pre = f"{pieces[1].upper()}: "
            else:
                pre = f"{pieces[0].title()}: "
        return f"{pre}{roman.numeral(self.num).lower()}"

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Scene":
        numeral = node.pieces[1]
        num = int(numeral) if numeral.isdigit() else roman.integer(numeral)
        setting = node.match.get("setting") or None
        parent = node.parent or node.act or node.scene
        if parent is None:
            raise ValueError(f"Can't build {cls.__name__!r} from node: {node}")
        return cls(
            index=node.index,
            text=node.match_text,
            num=num,
            act=parent,
            setting=setting,
        )


@typic.klass(unsafe_hash=True)
class Prologue(NodeMixin):
    """A representation of a single Prologue in a play.

    Notes:
        Prologues (and Epilogues) may have the body structure of either an Act or Scene.
    """

    type: ClassVar[NodeType] = NodeType.PROL
    index: int
    text: str
    setting: Optional[str] = None
    act: Optional[NodeID] = None
    body: "LogueBodyT" = typic.field(compare=False, hash=False, default=())  # type: ignore
    personae: "PersonaeIDT" = typic.field(compare=False, hash=False, default=())  # type: ignore
    as_act: bool = typic.field(init=False)  # type: ignore

    def __post_init__(self):
        # If the body conforms to the `ActBodyT` type, it should be treated as such.
        self.as_act = bool(
            self.body
            and isinstance(self.body[0], (Scene, Intermission, Epilogue, Prologue))
        )

    @typic.cached_property
    def id(self) -> NodeID:
        pre = f"{self.act}-" if self.act else ""
        return NodeID(f"{pre}{self.type.lower()}-{self.index}")

    @typic.cached_property
    def col(self):
        pre = self.act or ""
        if pre:
            pieces = self.act.split("-")
            if len(pieces) == 2:
                pre = f"{pieces[1].upper()}: "
            else:
                pre = f"{pieces[0].title()}: "
        return f"{pre}{self.type.title()}"

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Prologue":
        setting = node.match.get("setting") or None
        return cls(
            index=node.index, text=node.match_text, setting=setting, act=node.parent
        )


@typic.klass(unsafe_hash=True)
class Epilogue(Prologue):
    """A representation of a single Epilogue in a play.

    Notes:
        Epilogues (and Prologues) may have the body structure of either an Act or Scene.
    """

    type: ClassVar[NodeType] = NodeType.EPIL


@typic.klass(unsafe_hash=True)
class Intermission(NodeMixin):
    """A representation of an Intermission in a play."""

    type: ClassVar[NodeType] = NodeType.INTER
    index: int
    text: str
    act: NodeID

    @typic.cached_property
    def id(self) -> NodeID:
        return NodeID("intermission")

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Intermission":
        if not node.parent:
            raise ValueError(f"Can't build {cls.__name__!r} from node: {node}")
        return cls(index=node.index, text=node.match_text, act=node.parent)

    @typic.cached_property
    def col(self):
        return titleize(self.text)


@typic.klass(unsafe_hash=True)
class Persona(NodeMixin):
    """A representation of a single character in a Play."""

    type: ClassVar[NodeType] = NodeType.PERS
    index: int
    text: str
    name: str
    short: Optional[str] = None

    def __eq__(self, other) -> bool:
        return other.name == self.name if hasattr(other, "name") else False

    @typic.cached_property
    def id(self) -> NodeID:
        return NodeID(parameterize(self.name))

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Persona":
        return cls(
            index=node.index, text=node.match_text, name=titleize(node.match_text),
        )


@typic.klass(unsafe_hash=True)
class Entrance(NodeMixin):
    """A representation of an entrance for character(s) in a Scene."""

    type: ClassVar[NodeType] = NodeType.ENTER
    index: int
    text: str
    scene: NodeID
    personae: "PersonaeIDT" = ()

    @typic.cached_property
    def id(self) -> NodeID:
        return NodeID(f"{self.scene}-{self.type.lower()}-{self.index}")

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Entrance":
        if not node.parent:
            raise ValueError(f"Can't build {cls.__name__!r} from node: {node}")
        return cls(index=node.index, text=node.match_text, scene=node.parent)


@typic.klass(unsafe_hash=True)
class Exit(Entrance):
    """A representation of an exit for character(s) in a Scene."""

    type: ClassVar[NodeType] = NodeType.EXIT


@typic.klass(unsafe_hash=True)
class Action(NodeMixin):
    """A representation of a stage direction related to a specific character."""

    type: ClassVar[NodeType] = NodeType.ACTION
    action: str
    persona: NodeID
    scene: NodeID
    index: int

    @typic.cached_property
    def id(self) -> NodeID:
        return NodeID(f"{self.scene}-{self.persona}-{self.type.lower()}-{self.index}")

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Action":
        if not node.parent or not node.scene:
            raise ValueError(f"Can't build {cls.__name__!r} from node: {node}")
        return cls(
            action=node.match_text,
            persona=node.parent,
            scene=node.scene,
            index=node.index,
        )


@typic.klass(unsafe_hash=True)
class Direction(NodeMixin):
    """A representation of a stage direction."""

    type: ClassVar[NodeType] = NodeType.DIR
    action: str
    scene: NodeID
    index: int
    stop: bool = True

    @typic.cached_property
    def id(self) -> NodeID:
        return NodeID(f"{self.scene}-{self.type.lower()}-{self.index}")

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Direction":
        if not node.parent:
            raise ValueError(f"Can't build {cls.__name__!r} from node: {node}")
        return cls(action=node.match_text, scene=node.parent, index=node.index)


@typic.klass(unsafe_hash=True)
class Dialogue(NodeMixin):
    """A representation of a line of dialogue for a character in a scene."""

    type: ClassVar[NodeType] = NodeType.DIAL
    line: str
    persona: NodeID
    scene: NodeID
    index: int
    lineno: int
    linepart: int = 0

    @typic.cached_property
    def id(self) -> NodeID:
        return NodeID(
            f"{self.persona}-{self.type.lower()}-{self.lineno}-{self.linepart}"
        )

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Dialogue":
        if None in {node.parent, node.scene, node.lineno, node.linepart}:
            raise ValueError(f"Can't build {cls.__name__!r} from node: {node}")
        return cls(
            line=node.match_text,
            persona=node.parent,  # type: ignore
            scene=node.scene,  # type: ignore
            index=node.index,  # type: ignore
            lineno=node.lineno,  # type: ignore
            linepart=node.linepart,  # type: ignore
        )


@typic.klass(unsafe_hash=True)
class Speech(NodeMixin):
    """A representation of an unbroken piece of dialogue related to a single character."""

    type: ClassVar[NodeType] = NodeType.SPCH
    persona: NodeID
    scene: NodeID
    body: "SpeechBodyT"
    index: int

    def __post_init__(self):
        self.body = sort_body(self.body)

    @typic.cached_property
    def linerange(self) -> Tuple[int, int]:
        linenos = sorted((x.lineno for x in self.body if isinstance(x, Dialogue)))
        return linenos[0], linenos[-1]

    @typic.cached_property
    def num_lines(self) -> int:
        x, y = self.linerange
        # line count starts at 1
        # a range of 1 - 1 is 1 line
        return y - (x - 1)

    @typic.cached_property
    def id(self) -> NodeID:
        return NodeID(
            f"{self.scene}-{self.persona}-{self.type.lower()}-"
            f"{'{0}-{1}'.format(*self.linerange)}"
            f"-{self.index}"
        )


@typic.klass(unsafe_hash=True)
class Metadata(NodeMixin):
    """General information about a given play."""

    type: ClassVar[NodeType] = NodeType.META
    rights: str = "Creative Commons Non-Commercial Share Alike 3.0"
    language: str = "en-GB-emodeng"
    publisher: str = "Published w/ ❤️ using iambic - https://pypi.org/project/iambic"
    title: Optional[str] = None
    subtitle: Optional[str] = None
    edition: int = 1
    author: str = "William Shakespeare"
    editors: Tuple[str, ...] = ()
    tags: Tuple[str, ...] = ()

    @functools.lru_cache(1)
    def asmeta(self):  # pragma: nocover
        dikt = {
            "creator": [{"type": "author", "text": self.author}],
            "contributor": [{"type": "editor", "text": "MIT"}],
            "rights": self.rights,
            "language": self.language,
            "publisher": self.publisher,
            "subject": ["Shakespeare"],
        }
        if self.title:
            dikt["title"] = [
                {"type": "main", "text": self.title},
                {"type": "edition", "text": self.edition},
            ]
            if self.subtitle:
                dikt["title"].append({"type": "subtitle", "text": self.subtitle})
        dikt["contributor"].extend(
            ({"type": "editor", "text": x} for x in self.editors or [])
        )
        dikt["subject"].extend(self.tags or [])

        return dikt


@typic.klass(unsafe_hash=True)
class Play(NodeMixin):
    """A representation of a play in its entirety."""

    type: ClassVar[NodeType] = NodeType.PLAY
    body: "PlayBodyT" = ()
    personae: Tuple[Persona, ...] = ()
    meta: Metadata = typic.field(default_factory=Metadata)  # type: ignore

    def __post_init__(self):
        self.body = sort_body(self.body)

    @typic.cached_property
    def id(self) -> NodeID:
        return NodeID(parameterize(f"{self.meta.title}-{self.type.lower()}"))


@typic.klass(unsafe_hash=True)
class GenericNode(NodeMixin):
    """The root-object of a script.

    A script ``Node`` represents a single line of text in a script.
    """

    __resolver_map__: ClassVar[
        Mapping[NodeType, Type["ResolvedNodeT"]]
    ] = typic.FrozenDict(
        {
            NodeType.ACT: Act,
            NodeType.ACTION: Action,
            NodeType.DIAL: Dialogue,
            NodeType.DIR: Direction,
            NodeType.ENTER: Entrance,
            NodeType.EPIL: Epilogue,
            NodeType.EXIT: Exit,
            NodeType.INTER: Intermission,
            NodeType.META: Metadata,
            NodeType.PERS: Persona,
            NodeType.PLAY: Play,
            NodeType.PROL: Prologue,
            NodeType.SCENE: Scene,
            NodeType.SPCH: Speech,
        }
    )
    # Minimum data for a Node
    type: NodeType
    text: str
    index: int
    # Additional typic.fields which may be present
    lineno: Optional[int] = None
    linepart: Optional[int] = None
    # Given by text parser
    # If reading from JSON, we don't have/need this,
    # it will be provided inherently by the data-structure
    # on resolution-time.
    match: typic.FrozenDict = typic.field(default_factory=typic.FrozenDict)  # type: ignore
    parent: Optional[NodeID] = None
    act: Optional[NodeID] = None
    scene: Optional[NodeID] = None

    @typic.cached_property
    def resolved(self) -> "ResolvedNodeT":
        """Resolve a GenericNode into a typed, "resolved" Node.

        Raises
        ------
        TypeError
            If the NodeType provided has no resolved Node mapping.
        """
        if self.type in self.__resolver_map__:
            model: Type[ResolvedNodeT] = self.__resolver_map__[self.type]
            return model.from_node(self)  # type: ignore
        raise ValueError(
            f"Unrecognized node-type <{self.type}> for text <{self.text}>. "
            f"Valid types are: {tuple(self.__resolver_map__)}"
        )

    @typic.cached_property
    def pieces(self) -> List[str]:
        return self.match_text.split()

    @typic.cached_property
    def match_text(self) -> str:
        return self.match[self.type] if self.match else self.text


ResolvedNodeT = Union[
    Act,
    Scene,
    Prologue,
    Epilogue,
    Persona,
    Entrance,
    Exit,
    Action,
    Direction,
    Dialogue,
    Speech,
    Intermission,
]
SpeechNodeT = Union[Dialogue, Action, Direction]
SpeechBodyT = Tuple[SpeechNodeT, ...]
ActNodeT = Union[Scene, Intermission, Epilogue, Prologue]
ActBodyT = Tuple[ActNodeT, ...]
SceneNodeT = Union[Direction, Entrance, Exit, Speech]
SceneBodyT = Tuple[SceneNodeT, ...]
LogueBodyT = Union[ActBodyT, SceneBodyT]
PlayNodeT = Union[Act, Epilogue, Prologue]
PlayBodyT = Tuple[PlayNodeT, ...]
PersonaeIDT = Tuple[NodeID, ...]

_RESOLVABLE = set(GenericNode.__resolver_map__.values())


def node_deserializer(value: Any) -> Optional[ResolvedNodeT]:
    if type(value) in _RESOLVABLE or value is None:
        return value
    if isinstance(value, GenericNode):
        return value.resolved

    if not isinstance(value, Mapping):
        value = typic.transmute(dict, value)

    handler: Type[ResolvedNodeT] = GenericNode.__resolver_map__[value.pop("type")]
    resolved: ResolvedNodeT = typic.transmute(handler, value)  # type: ignore
    return resolved


def log_deserializer(value):
    return typic.protocol(Tuple[ResolvedNodeT, ...]).transmute(value)  # type: ignore


@functools.lru_cache(maxsize=None)
def isnodetype(obj: Type, *, __candidates=frozenset(_RESOLVABLE)) -> bool:
    is_valid = obj is GenericNode or (
        getattr(obj, "__origin__", None) is Union
        and {*getattr(obj, "__args__", ())}.issubset(__candidates)
    )
    return is_valid



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import functools
from collections import defaultdict
from itertools import chain
from typing import (
    DefaultDict,
    Dict,
    Optional,
    List,
    Set,
    Union,
    ValuesView,
    cast,
    Mapping,
    Iterable,
    Tuple,
)

import iambic.ast.node as ast

__all__ = ("Index",)


IndexKeyT = Union[ast.NodeType, str]
IndexInputT = Union[
    Mapping[str, ast.GenericNode], Iterable[Tuple[str, ast.GenericNode]]
]
SpeechMemberT = Union[ast.Action, ast.Dialogue, ast.Direction]


class Index(Dict[str, Union[ast.ResolvedNodeT, ast.GenericNode]]):
    """An ordered collection of nodes in a script.

    ``Parser.parse`` builds a list of :class:`GenericNode`, which can be resolved into the specific
    data-type for the :class:`NodeType` assigned.
    """

    def __init__(self, iterable: IndexInputT = None, **kwargs: ast.GenericNode):
        iterable = iterable or {}
        super().__init__(iterable, **kwargs)
        self.generic: Dict[str, ast.GenericNode] = {}
        self.__type_map: DefaultDict[
            ast.NodeType, Dict[str, ast.ResolvedNodeT]
        ] = defaultdict(dict)

    def __getitem__(self, item: IndexKeyT):
        if isinstance(item, ast.NodeType):
            return self.__type_map[item]
        return super().__getitem__(item)

    def __contains__(self, item: IndexKeyT) -> bool:  # type: ignore
        if isinstance(item, ast.NodeType):
            return item in self.__type_map
        return super().__contains__(item)

    def get(self, k: IndexKeyT, default=None):
        default = default or {}
        if k in self:
            return self[k]
        return default

    def get_values(self, k: ast.NodeType) -> ValuesView[ast.ResolvedNodeT]:
        return self.get(k, default={}).values()

    def add(self, node: ast.GenericNode):
        resolved = node.resolved
        self[resolved.id] = resolved
        self.generic[resolved.id] = node
        self.__type_map[resolved.type][resolved.id] = resolved

    @property
    def personae(self) -> ValuesView[ast.Persona]:
        return cast(ValuesView[ast.Persona], self.get_values(ast.NodeType.PERS))

    @property
    def acts(self) -> ValuesView[ast.Act]:
        return cast(ValuesView[ast.Act], self.get_values(ast.NodeType.ACT))

    @property
    def scenes(self) -> ValuesView[ast.Scene]:
        return cast(ValuesView[ast.Scene], self.get_values(ast.NodeType.SCENE))

    @property
    def prologues(self) -> ValuesView[ast.Prologue]:
        return cast(ValuesView[ast.Prologue], self.get_values(ast.NodeType.PROL))

    @property
    def epilogues(self) -> ValuesView[ast.Epilogue]:
        return cast(ValuesView[ast.Epilogue], self.get_values(ast.NodeType.EPIL))

    @property
    def dialogue(self) -> ValuesView[ast.Dialogue]:
        return cast(ValuesView[ast.Dialogue], self.get_values(ast.NodeType.DIAL))

    @property
    def directions(self) -> ValuesView[ast.Direction]:
        return cast(ValuesView[ast.Direction], self.get_values(ast.NodeType.DIR))

    @property
    def actions(self) -> ValuesView[ast.Action]:
        return cast(ValuesView[ast.Action], self.get_values(ast.NodeType.ACTION))

    @property
    def entrances(self) -> ValuesView[ast.Entrance]:
        return cast(ValuesView[ast.Entrance], self.get_values(ast.NodeType.ENTER))

    @property
    def exits(self) -> ValuesView[ast.Exit]:
        return cast(ValuesView[ast.Exit], self.get_values(ast.NodeType.EXIT))

    @property
    def intermission(self) -> Optional[ast.Intermission]:
        inter = {*self.get_values(ast.NodeType.INTER)}
        return cast(Optional[ast.Intermission], inter.pop() if inter else None)

    def resolve_presence(self):
        members: Iterable[Union[ast.Entrance, ast.Exit]] = chain(
            self.entrances, self.exits
        )
        personae = {x.text: x for x in self.personae}
        for member in members:
            present = (*(p.id for t, p in personae.items() if t in member.text),)
            member.personae = present

    def get_speeches(self) -> List[ast.Speech]:
        # Candidates for members of speeches.
        members: List[ast.SpeechNodeT] = sorted(
            chain(self.dialogue, self.directions, self.actions), key=ast.indexgetter
        )
        persona: Optional[ast.Persona] = None
        scene: ast.Scene = self[members[0].scene]
        speech: List[SpeechMemberT] = []
        speeches: List[ast.Speech] = []

        for node in members:
            # If we're in a new scene, we're definitely in a new speech.
            if node.scene != scene.id:
                if persona and scene and speech:
                    spch = ast.Speech(
                        persona.id, scene.id, tuple(speech), speech[0].index
                    )
                    speeches.append(spch)
                # Directions aren't directly associated to a persona,
                # so do not have the persona attr.
                # BUT they can occur within speeches, so we still have to track them all.
                persona = (
                    None if node.type == ast.NodeType.DIR else self[node.persona]  # type: ignore
                )
                scene, persona, speech = self[node.scene], persona, [node]
                continue
            # These puppies have a `persona` attr, so are easily associated to a speech.
            if isinstance(node, (ast.Action, ast.Dialogue)):
                # If the persona isn't set or has been reset.
                # See above for when persona can be set to null.
                if not persona:
                    persona = self[node.persona]
                    speech.append(node)
                    continue
                # Short-circuit the below if the personae match.
                if node.persona == persona.id:
                    speech.append(node)
                    continue
                # Otherwise, we have a new persona, meaning we have a new speech.
                spch = ast.Speech(persona.id, scene.id, tuple(speech), speech[0].index)
                speeches.append(spch)
                persona, scene, speech = self[node.persona], self[node.scene], [node]
                continue
            # If we've set a persona and we come across a Direction,
            # we can be reasonably sure this dir should be associated to this speech.
            if node.type == ast.NodeType.DIR and persona:
                speech.append(node)
        # Once we've exhausted all members, it's possible we have one speech left.
        if persona and scene and speech:
            spch = ast.Speech(persona.id, scene.id, tuple(speech), speech[0].index)
            speeches.append(spch)

        return speeches

    def filter_directions(self, speeches: Set[ast.Speech]) -> Set[ast.Direction]:
        speech_directions = {
            y for x in speeches for y in x.body if isinstance(y, ast.Direction)
        }
        return {*self.directions} - speech_directions

    def finalize_scenes(self) -> Iterable[ast.ActNodeT]:
        speeches = {*self.get_speeches()}
        directions = self.filter_directions(speeches)
        entrances = {*self.entrances}
        exits = {*self.exits}
        children = speeches | directions | entrances | exits
        scenes = []
        scene: Union[ast.Scene, ast.Epilogue, ast.Prologue]
        for scene in chain(self.scenes, self.epilogues, self.prologues):  # type: ignore
            child_nodes = {c for c in children if c.scene == scene.id}
            s: ast.Speech
            personae: Set[ast.NodeID] = {
                s.persona  # type: ignore
                for s in child_nodes
                if s.type == ast.NodeType.SPCH
            }
            if child_nodes:
                scene.body = ast.sort_body(child_nodes)
                scene.personae = (*personae,)
                scenes.append(scene)
                children -= child_nodes

        return scenes

    def finalize_acts(
        self, scenes: Iterable[ast.ActNodeT]
    ) -> List[Union[ast.Act, ast.Epilogue, ast.Prologue]]:
        scenes = {*scenes}
        logues: Set[Union[ast.Prologue, ast.Epilogue]] = {
            s
            for s in scenes
            if isinstance(s, (ast.Epilogue, ast.Prologue)) and not s.act
        }
        scenes -= logues
        intermission = self.intermission
        acts: List[ast.PlayNodeT] = [log for log in logues if not log.as_act]
        act: ast.PlayNodeT
        for act in chain(self.acts, (log for log in logues if log.as_act)):  # type: ignore
            children: Set[
                Union[ast.Scene, ast.Intermission, ast.Prologue, ast.Epilogue]
            ] = {s for s in scenes if s.act == act.id}
            if children:
                scenes -= children
                # Add the intermission if there is one.
                if intermission and intermission.act == act.id:
                    children.add(intermission)
                act.body = ast.sort_body(children)
                acts.append(act)
        return acts

    def resolve(self, meta: ast.Metadata = None) -> ast.Play:
        # Resolve the presence of personae in entrances/exits
        self.resolve_presence()
        # Build the Act-level trees
        scenes = self.finalize_scenes()
        # Build the Play-level trees
        acts = self.finalize_acts(scenes)
        # Put it all together
        play = ast.Play(
            ast.sort_body(acts), (*self.personae,), meta=(meta or ast.Metadata())
        )
        return play

    @staticmethod
    @functools.lru_cache(maxsize=2000)
    def node_sort(node: Union[ast.ResolvedNodeT, ast.GenericNode]):
        return node.index
